/*
 *************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *  
 *************************************************************************
 */

package org.eclipse.birt.data.engine.olap.data.impl.aggregation;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.birt.core.exception.BirtException;
import org.eclipse.birt.core.script.ScriptContext;
import org.eclipse.birt.core.script.ScriptExpression;
import org.eclipse.birt.data.engine.api.IBaseExpression;
import org.eclipse.birt.data.engine.core.DataException;
import org.eclipse.birt.data.engine.i18n.DataResourceHandle;
import org.eclipse.birt.data.engine.i18n.ResourceConstants;
import org.eclipse.birt.data.engine.olap.data.api.DimLevel;
import org.eclipse.birt.data.engine.olap.data.api.IAggregationResultSet;
import org.eclipse.birt.data.engine.olap.data.api.MeasureInfo;
import org.eclipse.birt.data.engine.olap.data.impl.DimColumn;
import org.eclipse.birt.data.engine.olap.data.impl.dimension.Member;
import org.eclipse.birt.data.engine.script.ScriptConstants;
import org.eclipse.birt.data.engine.script.ScriptEvalUtil;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;


/**
 * The data prepared for aggregation is from a <code>IAggregationResultSet</code> instance
 *
 */
public class DataSetFromAggregationResultSet implements IDataSet4Aggregation
{

	private static Logger logger = Logger.getLogger( DataSetFromAggregationResultSet.class.getName( ) );
	
	private IAggregationResultSet aggregationResultSet;
	
	//Dummy dimensions from dimLevles in  aggregationResultSet
	//the dimIndex, levelIndex are got from it
	DummyDimension[] dimensions;

	private int currentCursor = -1;

	private String addedMeasure;
	private IBaseExpression measureExpression;
	private Scriptable scope;
	private ScriptContext cx;
	
	/**
	 * 
	 * @param aggregationResultSet
	 * @param addedMeasure
	 * @param measureExpression
	 * @param scope
	 * @param cx
	 */
	public DataSetFromAggregationResultSet(
			IAggregationResultSet aggregationResultSet,
			String addedMeasure, IBaseExpression measureExpression,
			Scriptable scope, ScriptContext cx)
	{
		this.aggregationResultSet = aggregationResultSet;
		initDummyDimensions( );
		this.addedMeasure = addedMeasure;
		this.measureExpression = measureExpression;
		this.scope = scope;
		this.cx = cx;
		
		this.scope.put( ScriptConstants.DATA_BINDING_SCRIPTABLE, this.scope, new JSDataObject(  ));
		this.scope.put( ScriptConstants.DIMENSION_SCRIPTABLE, this.scope, new JSDimensionObject( ) );
	}

	private void initDummyDimensions( )
	{
		DimLevel[] dimLevels = aggregationResultSet.getAllLevels( );
		if ( dimLevels == null || dimLevels.length == 0 )
		{
			dimensions = new DummyDimension[0];
			return;
		}
		Map<String, Set<String>> dimNameToLevelNamesMap = new HashMap<String, Set<String>>( );
		for ( DimLevel dl : dimLevels )
		{
			String dimName = dl.getDimensionName( );
			Set<String> levelNames = dimNameToLevelNamesMap.get( dimName );
			if ( levelNames == null )
			{
				levelNames = new HashSet<String>( );
				dimNameToLevelNamesMap.put( dimName, levelNames );
			}
			levelNames.add( dl.getLevelName( ) );
		}
		dimensions = new DummyDimension[dimNameToLevelNamesMap.keySet( ).size( )];
		int i = 0;
		for ( String dimName : dimNameToLevelNamesMap.keySet( ) )
		{
			Set<String> levelNameSet = dimNameToLevelNamesMap.get( dimName );
			String[] levelNames = levelNameSet.toArray( new String[0] );
			DummyDimension dd = new DummyDimension( dimName, levelNames );
		
			int[] indexesInAggregationResultSet = new int[levelNames.length];
			for (int j = 0; j < indexesInAggregationResultSet.length; j++  )
			{
				DimLevel dl = new DimLevel(dimName, levelNames[j]);
				indexesInAggregationResultSet[j] = aggregationResultSet.getLevelIndex( dl );
			}
			dd.setDimLevelIndexesInAggregationResultSet( indexesInAggregationResultSet );
			dimensions[i++] = dd;
		}
	}

	private DummyDimension getDimension( String dimensionName )
	{
		for ( int i = 0; i < dimensions.length; i++ )
		{
			if ( dimensions[i].getDimensionName( ).equals( dimensionName ) )
			{
				return dimensions[i];
			}
		}
		return null;
	}

	private DummyDimension getDimension( int dimIndex )
	{
		return dimensions[dimIndex];
	}

	private int getDimensionIndex( String dimensionName )
	{
		for ( int i = 0; i < dimensions.length; i++ )
		{
			if ( dimensions[i].getDimensionName( ).equals( dimensionName ) )
			{
				return i;
			}
		}
		return -1;
	}

	private int getInnerIndexInAggregationResultSet( int dimIndex,
			int levelIndex )
	{
		DummyDimension dd = DataSetFromAggregationResultSet.this.getDimension( dimIndex );
		return dd.getIndexInAggregationResultSet( levelIndex );
	}

	public Object getMeasureValue( int measureIndex ) throws IOException, DataException
	{
		if ( measureIndex == aggregationResultSet.getAggregationCount( ) )
		{
			//added measure
			return ScriptEvalUtil.evalExpr( measureExpression,
					cx,
					this.scope,
					ScriptExpression.defaultID,
					0 );
		}
		return aggregationResultSet.getAggregationValue( measureIndex );
		
	}

	public Member getMember( int dimIndex, int levelIndex )
			throws DataException, IOException
	{
		int index = getInnerIndexInAggregationResultSet( dimIndex, levelIndex );
		return aggregationResultSet.getCurrentRow( ).getLevelMembers( )[index];
	}

	public MetaInfo getMetaInfo( )
	{
		return new IDataSet4Aggregation.MetaInfo( ) {

			public String[] getAttributeNames( int dimIndex, int levelIndex )
			{
				int index = getInnerIndexInAggregationResultSet( dimIndex,
						levelIndex );
				return aggregationResultSet.getAttributeNames( )[index];
			}

			public ColumnInfo getColumnInfo( DimColumn dimColumn )
					throws DataException
			{
				String dimensionName = dimColumn.getDimensionName( );
				String levelName = dimColumn.getLevelName( );
				String columnName = dimColumn.getColumnName( );

				int dimIndex = DataSetFromAggregationResultSet.this.getDimensionIndex( dimensionName );
				if ( dimIndex < 0 )
				{
					throw new DataException( DataResourceHandle.getInstance( )
							.getMessage( ResourceConstants.NONEXISTENT_DIMENSION )
							+ dimensionName );
				}

				DummyDimension dd = DataSetFromAggregationResultSet.this.getDimension( dimensionName );
				int levelIndex = dd.getLevelIndex( levelName );
				if ( levelIndex < 0 )
				{
					throw new DataException( DataResourceHandle.getInstance( )
							.getMessage( ResourceConstants.NONEXISTENT_LEVEL )
							+ "<" + dimensionName + " , " + levelName + ">" );   //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$
				}

				DimLevel dl = new DimLevel( dimensionName, levelName );
				int index = aggregationResultSet.getLevelIndex( dl );

				int columnIndex = -1;
				boolean isKey = false;
				String[] keys = aggregationResultSet.getKeyNames( )[index];
				for ( int i = 0; i < keys.length; i++ )
				{
					if ( keys[i].equals( columnName ) )
					{
						columnIndex = i;
						isKey = true;
						break;
					}
				}
				if ( !isKey )
				{
					String[] attrs = aggregationResultSet.getAttributeNames( )[index];
					for ( int i = 0; i < attrs.length; i++ )
					{
						if ( attrs[i].equals( columnName ) )
						{
							columnIndex = i;
							break;
						}
					}
				}
				if ( columnIndex < 0 )
				{
					throw new DataException( DataResourceHandle.getInstance( )
							.getMessage( ResourceConstants.NONEXISTENT_KEY_OR_ATTR )
							+ "<" //$NON-NLS-1$
							+ dimensionName
							+ " , " //$NON-NLS-1$
							+ levelName
							+ " , " //$NON-NLS-1$
							+ columnName + ">" ); //$NON-NLS-1$
				}
				return new IDataSet4Aggregation.ColumnInfo( dimIndex,
						levelIndex,
						columnIndex,
						isKey );
			}

			public int getDimensionIndex( String dimensionName )
			{
				return DataSetFromAggregationResultSet.this.getDimensionIndex( dimensionName );
			}

			public String[] getKeyNames( int dimIndex, int levelIndex )
			{
				int index = getInnerIndexInAggregationResultSet( dimIndex,
						levelIndex );
				return aggregationResultSet.getKeyNames( )[index];
			}

			public int getLevelIndex( String dimensionName, String levelName )
			{
				DummyDimension dd = DataSetFromAggregationResultSet.this.getDimension( dimensionName );
				if ( dd != null )
				{
					return dd.getLevelIndex( levelName );
				}
				return -1;
			}

			public int getMeasureIndex( String measureName ) throws IOException
			{
				int index = aggregationResultSet.getAggregationIndex( measureName );
				if ( index < 0 )
				{
					if ( addedMeasure.equals( measureName ))
					{
						return aggregationResultSet.getAggregationCount( );
					}
				}
				return index;
			}

			public MeasureInfo[] getMeasureInfos( ) throws IOException
			{
				int aggrCount = aggregationResultSet.getAggregationCount( );
				MeasureInfo[] mis = new MeasureInfo[aggrCount+1];
				for ( int i = 0; i < aggrCount; i++ )
				{
					MeasureInfo mi = new MeasureInfo( aggregationResultSet.getAggregationName( i ),
							aggregationResultSet.getAggregationDataType( i ) );
					mis[i] = mi;
				}
				mis[aggrCount] = new MeasureInfo( addedMeasure, measureExpression.getDataType( ) );
				return mis;
			}

		};
	}

	public boolean next( ) throws DataException, IOException
	{
		if ( aggregationResultSet.length( ) <= 0 )
		{
			return false;
		}
		currentCursor++;
		if ( currentCursor < aggregationResultSet.length( ) )
		{
			aggregationResultSet.seek( currentCursor );
			return true;
		}
		return false;
	}

	static class DummyDimension
	{
		String dimensionName;
		
		//the level names in this dimension
		String[] levelNames;
		
		//the corresponding DimLevel index in aggregationResultSet
		//one levelName in levelNames has a corresponding dimLevelIndex from <code>AggregationResultSet.getLevelIndex( DimLevel dl )</code>
		int[] dimLevelIndexesInAggregationResultSet;
		

		public DummyDimension( String dimensionName, String[] levelNames )
		{
			this.dimensionName = dimensionName;
			this.levelNames = levelNames;
		}
		
		public void setDimLevelIndexesInAggregationResultSet(int[] indexes)
		{
			dimLevelIndexesInAggregationResultSet = indexes;
		}
		
		public int getIndexInAggregationResultSet(int levelIndex)
		{
			return dimLevelIndexesInAggregationResultSet[levelIndex];
		}

		public String getLevelName( int index )
		{
			return levelNames[index];
		}

		public int getLevelIndex( String name )
		{
			for ( int i = 0; i < levelNames.length; i++ )
			{
				if ( levelNames[i].equals( name ) )
				{
					return i;
				}
			}
			return -1;
		}

		public String getDimensionName( )
		{
			return dimensionName;
		}
	}
	
	public void close( ) throws BirtException, IOException
	{
		
	}
	
	@SuppressWarnings("serial")
	private class JSDataObject extends ScriptableObject
	{
		/*
		 * (non-Javadoc)
		 * @see org.mozilla.javascript.ScriptableObject#get(java.lang.String, org.mozilla.javascript.Scriptable)
		 */
		public Object get( String arg0, Scriptable scope )
		{
			DataSetFromAggregationResultSet dsfars = DataSetFromAggregationResultSet.this;
			try
			{
				int index = dsfars.getMetaInfo( ).getMeasureIndex( arg0 );
				if ( index < 0 )
				{
					throw Context.reportRuntimeError( ResourceConstants.INVALID_NEST_AGGREGATION_EXPRESSION );
				}
				else
				{
					return dsfars.getMeasureValue( index );
				}
			}
			catch ( IOException e )
			{
				logger.log( Level.WARNING, e.getLocalizedMessage( ), e ); 
				throw Context.reportRuntimeError( e.getLocalizedMessage( ) );
			}
			catch ( DataException e )
			{
				logger.log( Level.WARNING, e.getLocalizedMessage( ), e );
				throw Context.reportRuntimeError( e.getLocalizedMessage( ) );
			}
		}
		
		/*
		 * (non-Javadoc)
		 * @see org.mozilla.javascript.ScriptableObject#getClassName()
		 */
		public String getClassName( )
		{
			return JSDataObject.class.getName( );
		}

	}
	
	@SuppressWarnings("serial")
	private class JSDimensionObject extends ScriptableObject
	{

		/* (non-Javadoc)
		 * @see org.mozilla.javascript.ScriptableObject#get(java.lang.String, org.mozilla.javascript.Scriptable)
		 */
		@Override
		public Object get( String name, Scriptable start )
		{
			DataSetFromAggregationResultSet dsfars = DataSetFromAggregationResultSet.this;
			if ( dsfars.getMetaInfo( ).getDimensionIndex( name ) < 0 )
			{
				throw Context.reportRuntimeError( ResourceConstants.INVALID_NEST_AGGREGATION_EXPRESSION );
			}
			return new JSLevelObject( name );
		}

		@Override
		public String getClassName( )
		{
			return JSDimensionObject.class.getName( );
		}
		
	}
	
	@SuppressWarnings("serial")
	private class JSLevelObject extends ScriptableObject
	{
		private String dimensionName;
		
		public JSLevelObject( String dimensionName )
		{
			this.dimensionName = dimensionName;
		}

		/* (non-Javadoc)
		 * @see org.mozilla.javascript.ScriptableObject#get(java.lang.String, org.mozilla.javascript.Scriptable)
		 */
		@Override
		public Object get( String name, Scriptable start )
		{
			DataSetFromAggregationResultSet dsfars = DataSetFromAggregationResultSet.this;
			int dimIndex = dsfars.getDimensionIndex( dimensionName );
			int levelIndex = dsfars.getMetaInfo( ).getLevelIndex( dimensionName, name );
			if ( levelIndex < 0  )
			{
				throw Context.reportRuntimeError( ResourceConstants.INVALID_NEST_AGGREGATION_EXPRESSION );
			}
			return new JSMemberObject( dimIndex, levelIndex );
		}

		@Override
		public String getClassName( )
		{
			return JSLevelObject.class.getName( );
		}
		
	}
	
	@SuppressWarnings("serial")
	private class JSMemberObject extends ScriptableObject
	{
		private int dimensionIndex;
		private int levelIndex;
		
		public JSMemberObject( int dimensionIndex, int levelIndex )
		{
			this.dimensionIndex = dimensionIndex;
			this.levelIndex = levelIndex;
		}
		
		

		/* (non-Javadoc)
		 * @see org.mozilla.javascript.ScriptableObject#getDefaultValue(java.lang.Class)
		 */
		@SuppressWarnings("unchecked")
		@Override
		public Object getDefaultValue( Class typeHint )
		{
			DataSetFromAggregationResultSet dsfars = DataSetFromAggregationResultSet.this;
			try
			{
				Member m = dsfars.getMember( dimensionIndex, levelIndex );
				
				//default value is its first key value
				return m.getKeyValues( )[0];
			}
			catch ( DataException e )
			{
				logger.log( Level.WARNING, e.getLocalizedMessage( ), e ); 
				throw Context.reportRuntimeError( e.getLocalizedMessage( ) );
			}
			catch ( IOException e )
			{
				logger.log( Level.WARNING, e.getLocalizedMessage( ), e ); 
				throw Context.reportRuntimeError( e.getLocalizedMessage( ) );
			}
		}



		/* (non-Javadoc)
		 * @see org.mozilla.javascript.ScriptableObject#get(java.lang.String, org.mozilla.javascript.Scriptable)
		 */
		@Override
		public Object get( String name, Scriptable start )
		{
			DataSetFromAggregationResultSet dsfars = DataSetFromAggregationResultSet.this;
			try
			{
				Member m = dsfars.getMember( dimensionIndex, levelIndex );
				String[] keyNames = dsfars.getMetaInfo( ).getKeyNames( dimensionIndex, levelIndex );
				for ( int i=0; i<keyNames.length; i++)
				{
					if ( name.equals( keyNames[i] ))
					{
						return m.getKeyValues( )[i];
					}
				}
				String[] attrNames = dsfars.getMetaInfo( ).getAttributeNames( dimensionIndex, levelIndex );
				for ( int i=0; i<attrNames.length; i++)
				{
					if ( name.equals( attrNames[i] ))
					{
						return m.getAttributes( )[i];
					}
				}
				throw Context.reportRuntimeError( ResourceConstants.INVALID_NEST_AGGREGATION_EXPRESSION );
			}
			catch ( DataException e )
			{
				logger.log( Level.WARNING, e.getLocalizedMessage( ), e ); 
				throw Context.reportRuntimeError( e.getLocalizedMessage( ) );
			}
			catch ( IOException e )
			{
				logger.log( Level.WARNING, e.getLocalizedMessage( ), e ); 
				throw Context.reportRuntimeError( e.getLocalizedMessage( ) );
			}
		}

		@Override
		public String getClassName( )
		{
			return JSMemberObject.class.getName( );
		}
		
	}
}
